@@ -5,20 +5,21 @@
 # This file is part of weboob.
 #
 # weboob is free software: you can redistribute it and/or modify
-# it under the terms of the GNU Affero General Public License as published by
+# it under the terms of the GNU Lesser General Public License as published by
 # the Free Software Foundation, either version 3 of the License, or
 # (at your option) any later version.
 #
 # weboob is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
-# GNU Affero General Public License for more details.
+# GNU Lesser General Public License for more details.
 #
-# You should have received a copy of the GNU Affero General Public License
+# You should have received a copy of the GNU Lesser General Public License
 # along with weboob. If not, see <http://www.gnu.org/licenses/>.
 
 from __future__ import print_function
 
+from contextlib import contextmanager
 import datetime
 import uuid
 from dateutil.relativedelta import relativedelta
@@ -27,14 +28,25 @@ from decimal import Decimal, InvalidOperation
 
 from weboob.browser.browsers import APIBrowser
 from weboob.browser.profiles import Weboob
-from weboob.exceptions import BrowserHTTPError
-from weboob.capabilities.base import empty
-from weboob.capabilities.bank import CapBank, Account, Transaction, CapBankTransfer, \
-                                     Transfer, TransferStep, Recipient, AddRecipientStep
+from weboob.exceptions import BrowserHTTPError, CaptchaQuestion
+from weboob.core.bcall import CallErrors
+from weboob.capabilities.base import empty, find_object
+from weboob.capabilities.bank import (
+    Account, Transaction,
+    Transfer, TransferStep, Recipient, AddRecipientStep,
+    CapBank, CapBankTransfer, CapBankWealth, CapBankPockets,
+    TransferInvalidLabel, TransferInvalidAmount, TransferInvalidDate,
+    TransferInvalidEmitter, TransferInvalidRecipient,
+)
+from weboob.capabilities.captcha import exception_to_job
 from weboob.capabilities.contact import CapContact, Advisor
 from weboob.capabilities.profile import CapProfile
 from weboob.tools.application.repl import ReplApplication, defaultcount
+from weboob.tools.application.captcha import CaptchaMixin
 from weboob.tools.application.formatters.iformatter import IFormatter, PrettyFormatter
+from weboob.tools.compat import getproxies
+from weboob.tools.log import getLogger
+from weboob.tools.misc import to_unicode
 
 
 __all__ = ['Boobank']
@@ -65,11 +77,14 @@ class OfxFormatter(IFormatter):
 
     balance = Decimal(0)
     coming = Decimal(0)
+    account_type = ''
+    seen = set()
 
     def start_format(self, **kwargs):
         account = kwargs['account']
         self.balance = account.balance
         self.coming = account.coming
+        self.account_type = account.type
 
         self.output(u'OFXHEADER:100')
         self.output(u'DATA:OFXSGML')
@@ -82,16 +97,28 @@ class OfxFormatter(IFormatter):
         self.output(u'NEWFILEUID:%s\n' % uuid.uuid1())
         self.output(u'<OFX><SIGNONMSGSRSV1><SONRS><STATUS><CODE>0<SEVERITY>INFO</STATUS>')
         self.output(u'<DTSERVER>%s113942<LANGUAGE>ENG</SONRS></SIGNONMSGSRSV1>' % datetime.date.today().strftime('%Y%m%d'))
-        self.output(u'<BANKMSGSRSV1><STMTTRNRS><TRNUID>%s' % uuid.uuid1())
-        self.output(u'<STATUS><CODE>0<SEVERITY>INFO</STATUS><CLTCOOKIE>null<STMTRS>')
-        self.output(u'<CURDEF>%s<BANKACCTFROM>' % (account.currency or 'EUR'))
-        self.output(u'<BANKID>null')
-        self.output(u'<BRANCHID>null')
-        self.output(u'<ACCTID>%s' % account.id)
-        account_type = self.TYPES_ACCTS.get(account.type, 'CHECKING')
-        self.output(u'<ACCTTYPE>%s' % account_type)
-        self.output(u'<ACCTKEY>null')
-        self.output('</BANKACCTFROM>')
+
+        if self.account_type == Account.TYPE_CARD:
+            self.output(u'<CREDITCARDMSGSRSV1><CCSTMTTRNRS><TRNUID>%s' % uuid.uuid1())
+            self.output(u'<STATUS><CODE>0<SEVERITY>INFO</STATUS><CLTCOOKIE>null<CCSTMTRS>')
+            self.output(u'<CURDEF>%s<CCACCTFROM>' % (account.currency or 'EUR'))
+            self.output(u'<BANKID>null')
+            self.output(u'<BRANCHID>null')
+            self.output(u'<ACCTID>%s' % account.id)
+            self.output(u'<ACCTTYPE>%s' % self.account_type)
+            self.output(u'<ACCTKEY>null')
+            self.output('</CCACCTFROM>')
+        else:
+            self.output(u'<BANKMSGSRSV1><STMTTRNRS><TRNUID>%s' % uuid.uuid1())
+            self.output(u'<STATUS><CODE>0<SEVERITY>INFO</STATUS><CLTCOOKIE>null<STMTRS>')
+            self.output(u'<CURDEF>%s<BANKACCTFROM>' % (account.currency or 'EUR'))
+            self.output(u'<BANKID>null')
+            self.output(u'<BRANCHID>null')
+            self.output(u'<ACCTID>%s' % account.id)
+            self.output(u'<ACCTTYPE>%s' % self.account_type)
+            self.output(u'<ACCTKEY>null')
+            self.output('</BANKACCTFROM>')
+
         self.output(u'<BANKTRANLIST>')
         self.output(u'<DTSTART>%s' % datetime.date.today().strftime('%Y%m%d'))
         self.output(u'<DTEND>%s' % datetime.date.today().strftime('%Y%m%d'))
@@ -110,7 +137,7 @@ class OfxFormatter(IFormatter):
         if obj.rdate:
             result += u'<DTUSER>%s\n' % obj.rdate.strftime('%Y%m%d')
         result += u'<TRNAMT>%s\n' % obj.amount
-        result += u'<FITID>%s\n' % obj.unique_id()
+        result += u'<FITID>%s\n' % obj.unique_id(self.seen)
 
         if hasattr(obj, 'label') and not empty(obj.label):
             result += u'<NAME>%s\n' % obj.label.replace('&', '&amp;')
@@ -132,9 +159,12 @@ class OfxFormatter(IFormatter):
             self.output(u'<AVAILBAL><BALAMT>%s' % (self.balance + self.coming))
         except TypeError:
             self.output(u'<AVAILBAL><BALAMT>%s' % self.balance)
-
         self.output(u'<DTASOF>%s</AVAILBAL>' % datetime.date.today().strftime('%Y%m%d'))
-        self.output(u'</STMTRS></STMTTRNRS></BANKMSGSRSV1></OFX>')
+
+        if self.account_type == Account.TYPE_CARD:
+            self.output(u'</CCSTMTRS></CCSTMTTRNRS></CREDITCARDMSGSRSV1></OFX>')
+        else:
+            self.output(u'</STMTRS></STMTTRNRS></BANKMSGSRSV1></OFX>')
 
 
 class QifFormatter(IFormatter):
@@ -207,17 +237,25 @@ class TransactionsFormatter(IFormatter):
 
 
 class TransferFormatter(IFormatter):
-    MANDATORY_FIELDS = ('id', 'date', 'origin', 'recipient', 'amount')
-    DISPLAYED_FIELDS = ('reason', )
+    MANDATORY_FIELDS = ('id', 'exec_date', 'account_label', 'recipient_label', 'amount')
+    DISPLAYED_FIELDS = ('label', 'account_iban', 'recipient_iban', 'currency')
 
     def format_obj(self, obj, alias):
         result = u'------- Transfer %s -------\n' % obj.fullid
-        result += u'Date:       %s\n' % obj.date
-        result += u'Origin:     %s\n' % obj.origin
-        result += u'Recipient:  %s\n' % obj.recipient
-        result += u'Amount:     %.2f\n' % obj.amount
-        if obj.reason:
-            result += u'Reason:     %s\n' % obj.reason
+        result += u'Date:       %s\n' % obj.exec_date
+        if obj.account_iban:
+            result += u'Origin:     %s (%s)\n' % (obj.account_label, obj.account_iban)
+        else:
+            result += u'Origin:     %s\n' % obj.account_label
+
+        if obj.recipient_iban:
+            result += u'Recipient:  %s (%s)\n' % (obj.recipient_label, obj.recipient_iban)
+        else:
+            result += u'Recipient:  %s\n' % obj.recipient_label
+
+        result += u'Amount:     %.2f %s\n' % (obj.amount, obj.currency or '')
+        if obj.label:
+            result += u'Label:      %s\n' % obj.label
         return result
 
 
@@ -291,11 +329,15 @@ class InvestmentFormatter(IFormatter):
 
 class RecipientListFormatter(PrettyFormatter):
     MANDATORY_FIELDS = ('id', 'label')
+    DISPLAYED_FIELDS = ('iban', 'bank_name')
 
     def start_format(self, **kwargs):
         self.output('Available recipients:')
 
     def get_title(self, obj):
+        details = ' - '.join(filter(None, (obj.iban, obj.bank_name)))
+        if details:
+            return '%s (%s)' % (obj.label, details)
         return obj.label
 
 
@@ -389,11 +431,11 @@ class AccountListFormatter(IFormatter):
         self.totals.clear()
 
 
-class Boobank(ReplApplication):
+class Boobank(CaptchaMixin, ReplApplication):
     APPNAME = 'boobank'
-    VERSION = '1.3'
+    VERSION = '1.5'
     COPYRIGHT = 'Copyright(C) 2010-YEAR Romain Bignon, Christophe Benz'
-    CAPS = CapBank, CapBankTransfer
+    CAPS = CapBank
     DESCRIPTION = "Console application allowing to list your bank accounts and get their balance, " \
                   "display accounts history and coming bank operations, and transfer money from an account to " \
                   "another (if available)."
@@ -411,6 +453,7 @@ class Boobank(ReplApplication):
     DEFAULT_FORMATTER = 'table'
     COMMANDS_FORMATTERS = {'ls':          'account_list',
                            'list':        'account_list',
+                           'recipients':  'recipient_list',
                            'transfer':    'transfer',
                            'history':     'ops_list',
                            'coming':      'ops_list',
@@ -438,9 +481,31 @@ class Boobank(ReplApplication):
                 v = self.ask(field)
                 if v:
                     params[field.id] = v
-            next(iter(self.do('add_recipient', error.recipient, **params)))
+            try:
+                next(iter(self.do('add_recipient', error.recipient, **params)))
+            except CallErrors as e:
+                self.bcall_errors_handler(e)
+        elif isinstance(error, TransferInvalidAmount):
+            print(u'Error(%s): %s' % (backend.name, to_unicode(error) or 'The transfer amount is invalid'), file=self.stderr)
+        elif isinstance(error, TransferInvalidLabel):
+            print(u'Error(%s): %s' % (backend.name, to_unicode(error) or 'The transfer label is invalid'), file=self.stderr)
+        elif isinstance(error, TransferInvalidEmitter):
+            print(u'Error(%s): %s' % (backend.name, to_unicode(error) or 'The transfer emitter is invalid'), file=self.stderr)
+        elif isinstance(error, TransferInvalidRecipient):
+            print(u'Error(%s): %s' % (backend.name, to_unicode(error) or 'The transfer recipient is invalid'), file=self.stderr)
+        elif isinstance(error, TransferInvalidDate):
+            print(u'Error(%s): %s' % (backend.name, to_unicode(error) or 'The transfer execution date is invalid'), file=self.stderr)
+        elif isinstance(error, CaptchaQuestion):
+            if not self.captcha_weboob.count_backends():
+                print('Error(%s): Site requires solving a CAPTCHA but no CapCaptchaSolver backends were configured' % backend.name, file=self.stderr)
+                return False
+
+            print('Info(%s): Encountered CAPTCHA, please wait for its resolution, it can take dozens of seconds' % backend.name, file=self.stderr)
+            job = exception_to_job(error)
+            self.solve_captcha(job, backend)
+            return False
         else:
-            return ReplApplication.bcall_error_handler(self, backend, error, backtrace)
+            return super(Boobank, self).bcall_error_handler(backend, error, backtrace)
 
     def load_default_backends(self):
         self.load_backends(CapBank, storage=self.create_storage())
@@ -532,84 +597,167 @@ class Boobank(ReplApplication):
 
         Add a recipient to a backend.
         """
-        iban, label = self.parse_command_args(line, 2, 2)
+        if len(self.enabled_backends) > 1:
+            print('Error: select a single backend to add a recipient (Hint: try the command "backends only")', file=self.stderr)
+            return 1
+        iban, label, origin_account_id = self.parse_command_args(line, 3, 2)
         recipient = Recipient()
         recipient.iban = iban
         recipient.label = label
+        recipient.origin_account_id = origin_account_id
         next(iter(self.do('add_recipient', recipient)))
 
-    def do_transfer(self, line):
+    def do_recipients(self, line):
         """
-        transfer ACCOUNT [RECIPIENT AMOUNT [REASON]]
+        recipients ACCOUNT
 
-        Make a transfer beetwen two account
-        - ACCOUNT    the source account
-        - RECIPIENT  the recipient
-        - AMOUNT     amount to transfer
-        - REASON     reason of transfer
-
-        If you give only the ACCOUNT parameter, it lists all the
-        available recipients for this account.
+        List recipients of ACCOUNT
         """
-        id_from, id_to, amount, reason = self.parse_command_args(line, 4, 1)
+        id_from, = self.parse_command_args(line, 1, 1)
 
         account = self.get_object(id_from, 'get_account', [])
         if not account:
             print('Error: account %s not found' % id_from, file=self.stderr)
             return 1
 
-        if not id_to:
-            self.objects = []
-            self.set_formatter('table')
-            self.set_formatter_header(u'Available recipients')
+        self.objects = []
 
-            self.start_format()
-            for recipient in self.do('iter_transfer_recipients', account.id, backends=account.backend, caps=CapBankTransfer):
-                self.cached_format(recipient)
-            return 0
+        self.start_format()
+        for recipient in self.do('iter_transfer_recipients', account.id, backends=account.backend, caps=CapBankTransfer):
+            self.cached_format(recipient)
 
-        id_to, backend_name_to = self.parse_id(id_to)
+    @contextmanager
+    def use_cmd_formatter(self, cmd):
+        self.set_formatter(self.commands_formatters.get(cmd, self.DEFAULT_FORMATTER))
+        try:
+            yield
+        finally:
+            self.flush()
 
-        if account.backend != backend_name_to:
-            print("Transfer between different backends is not implemented", file=self.stderr)
-            return 4
+    def _build_transfer(self, line):
+        if self.interactive:
+            id_from, id_to, amount, reason, exec_date = self.parse_command_args(line, 5, 0)
+        else:
+            id_from, id_to, amount, reason, exec_date = self.parse_command_args(line, 5, 3)
 
+        missing = not bool(id_from and id_to and amount)
+
+        if id_from:
+            account = self.get_object(id_from, 'get_account', [])
+            id_from = account.id
+            if not account:
+                print('Error: account %s not found' % id_from, file=self.stderr)
+                return
+        else:
+            with self.use_cmd_formatter('list'):
+                self.do_ls('')
+            id_from = self.ask('Transfer money from account', default='')
+            if not id_from:
+                return
+            id_from, backend = self.parse_id(id_from)
+
+            account = find_object(self.objects, fullid='%s@%s' % (id_from, backend))
+            if not account:
+                return
+            id_from = account.id
+
+        if id_to:
+            id_to, backend_name_to = self.parse_id(id_to)
+            if account.backend != backend_name_to:
+                print("Transfer between different backends is not implemented", file=self.stderr)
+                return
+            rcpts = self.do('iter_transfer_recipients', id_from, backends=account.backend)
+            rcpt = find_object(rcpts, id=id_to)
+        else:
+            with self.use_cmd_formatter('recipients'):
+                self.do_recipients(account.fullid)
+            id_to = self.ask('Transfer money to recipient', default='')
+            if not id_to:
+                return
+            id_to, backend = self.parse_id(id_to)
+
+            rcpt = find_object(self.objects, fullid='%s@%s' % (id_to, backend))
+            if not rcpt:
+                return
+
+        if not amount:
+            amount = self.ask('Amount to transfer', default='', regexp=r'\d+(?:\.\d*)?')
         try:
             amount = Decimal(amount)
         except (TypeError, ValueError, InvalidOperation):
             print('Error: please give a decimal amount to transfer', file=self.stderr)
-            return 2
+            return
+        if amount <= 0:
+            print('Error: transfer amount must be strictly positive', file=self.stderr)
+            return
 
-        exec_date = datetime.date.today()
+        if missing:
+            reason = self.ask('Label of the transfer (seen by the recipient)', default='')
+            exec_date = self.ask('Execution date of the transfer (YYYY-MM-DD format, empty for today)', default='')
 
-        if self.interactive:
+        today = datetime.date.today()
+        if exec_date:
+            try:
+                exec_date = datetime.datetime.strptime(exec_date, '%Y-%m-%d').date()
+            except ValueError:
+                print('Error: execution date must be valid and in YYYY-MM-DD format', file=self.stderr)
+                return
+            if exec_date < today:
+                print('Error: execution date cannot be in the past', file=self.stderr)
+                return
+        else:
+            exec_date = today
+
+        transfer = Transfer()
+        transfer.backend = account.backend
+        transfer.account_id = account.id
+        transfer.account_label = account.label
+        transfer.account_iban = account.iban
+        transfer.recipient_id = id_to
+        if rcpt:
             # Try to find the recipient label. It can be missing from
             # recipients list, for example for banks which allow transfers to
             # arbitrary recipients.
-            to = id_to
-            for recipient in self.do('iter_transfer_recipients', account.id, backends=account.backend, caps=CapBankTransfer):
-                if recipient.id == id_to:
-                    to = recipient.label
-                    break
+            transfer.recipient_label = rcpt.label
+            transfer.recipient_iban = rcpt.iban
+        transfer.amount = amount
+        transfer.label = reason or u''
+        transfer.exec_date = exec_date
+
+        return transfer
+
+    def do_transfer(self, line):
+        """
+        transfer [ACCOUNT RECIPIENT AMOUNT [LABEL [EXEC_DATE]]]
+
+        Make a transfer beetwen two accounts
+        - ACCOUNT    the source account
+        - RECIPIENT  the recipient
+        - AMOUNT     amount to transfer
+        - LABEL      label of transfer
+        - EXEC_DATE  date when to execute the transfer
+        """
+
+        transfer = self._build_transfer(line)
+        if transfer is None:
+            return 1
+
+        if self.interactive:
+            with self.use_cmd_formatter('transfer'):
+                self.start_format()
+                self.cached_format(transfer)
 
-            print('Amount: %s%s' % (amount, account.currency_text))
-            print('From:   %s' % account.label)
-            print('To:     %s' % to)
-            print('Reason: %s' % (reason or ''))
-            print('Date:   %s' % exec_date)
             if not self.ask('Are you sure to do this transfer?', default=True):
                 return
 
+        # only keep basic fields because most modules don't handle others
+        transfer.account_label = None
+        transfer.account_iban = None
+        transfer.recipient_label = None
+        transfer.recipient_iban = None
+
         self.start_format()
-
-        transfer = Transfer()
-        transfer.account_id = account.id
-        transfer.recipient_id = id_to
-        transfer.amount = amount
-        transfer.label = reason or ''
-        transfer.exec_date = exec_date
-
-        next(iter(self.do('transfer', transfer, backends=account.backend)))
+        next(iter(self.do('transfer', transfer, backends=transfer.backend)))
 
     def show_wealth(self, command, id):
         account = self.get_object(id, 'get_account', [])
@@ -617,8 +765,13 @@ class Boobank(ReplApplication):
             print('Error: account "%s" not found (Hint: try the command "list")' % id, file=self.stderr)
             return 2
 
+        caps = {
+            'iter_investment': CapBankWealth,
+            'iter_pocket': CapBankPockets,
+        }
+
         self.start_format()
-        for el in self.do(command, account, backends=account.backend):
+        for el in self.do(command, account, backends=account.backend, caps=caps[command]):
             self.format(el)
 
     def do_investment(self, id):
@@ -651,7 +804,9 @@ class Boobank(ReplApplication):
         """
         username, password = self.parse_command_args(line, 2, 2)
 
-        client = APIBrowser(baseurl='https://budgea.biapi.pro/2.0/')
+        client = APIBrowser(baseurl='https://budgea.biapi.pro/2.0/',
+                            proxy=getproxies(),
+                            logger=getLogger('apibrowser', self.logger))
         client.set_profile(Weboob(self.VERSION))
         client.TIMEOUT = 60
         try:
