@@ -5,16 +5,16 @@
 # This file is part of weboob.
 #
 # weboob is free software: you can redistribute it and/or modify
-# it under the terms of the GNU Affero General Public License as published by
+# it under the terms of the GNU Lesser General Public License as published by
 # the Free Software Foundation, either version 3 of the License, or
 # (at your option) any later version.
 #
 # weboob is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
-# GNU Affero General Public License for more details.
+# GNU Lesser General Public License for more details.
 #
-# You should have received a copy of the GNU Affero General Public License
+# You should have received a copy of the GNU Lesser General Public License
 # along with weboob. If not, see <http://www.gnu.org/licenses/>.
 
 
@@ -248,7 +248,6 @@ class BackendCfg(QDialog):
         self.is_enabling += 1
 
         backend_name = item.text(0)
-        module_name = item.text(1)
         if item.checkState(0) == Qt.Checked:
             self.to_load.add(backend_name)
             enabled = 'true'
@@ -260,7 +259,7 @@ class BackendCfg(QDialog):
                 pass
             enabled = 'false'
 
-        self.weboob.backends_config.edit_backend(backend_name, module_name, {'_enabled': enabled})
+        self.weboob.backends_config.edit_backend(backend_name, {'_enabled': enabled})
 
     @Slot(QTreeWidgetItem, int)
     def backendClicked(self, item, col):
